# C++ cp Command Line Program

This project implements an enhanced version of the Unix `cp` command in C++.

---

## Overview

The program is a terminal-based application written in C++ that copies files and directories with standard and unique features. It uses the `std::experimental::filesystem` library and is developed on `csctcloud.uwe.ac.uk`.

### Features
- **Basic File Copy**: Copies a single file to a destination.
- **Standard Options**:
  - `-r`: Recursive copy for directories.
  - `-i`: Interactive mode (prompts before overwriting).
  - `-v`: Verbose mode (displays actions).
  - `-p`: Preserves file attributes (permissions and timestamps).
  - `-u`: Update mode (copies only if source is newer).
  - `-n`: Dry-run mode (simulates actions without copying).
  - `-h`: Displays help message.
- **Unique Features**:
  - **Wildcard Support**: Copies multiple files matching a pattern (e.g., `*.txt`).
  - **Progress Indicator**: Shows real-time progress during directory copies.

### Files
- `cp.hpp`: Header file with function declarations.
- `cp.cpp`: Implementation of file and directory copying.
- `main.cpp`: Command-line interface and argument parsing.
- `README.md`: The documentation.

---

## Requirements

- **Compiler**: GCC 9.4.0 or later with C++17 support.
- **Library**: `libstdc++fs` (linked with `-lstdc++fs`).
- **Environment**: Developed on `csctcloud.uwe.ac.uk`.
- **GitLab**: Project hosted in a GitLab repository.

---

## Build Instructions

1. **Clone the Repository**:
   ```bash
   git clone https://gitlab.uwe.ac.uk/m3-alabdulla/cp-command-line-program

2. **Navigate to Directory**:
   ```bash
   cd cp-command-line-program

3. **Compile the Program**:
   ```bash
   g++ -std=c++17 main.cpp cp.cpp -o cp -lstdc++fs
   -  std=c++17: Enables C++17 features.
   -  lstdc++fs: Links the filesystem library.

4. **Verify Executable**:
   ```bash
   ls  # Should show 'cp' among the files

## Usage

```
./cp [options] source_file... destination or ../cp # depend upon where exe file is saved
```

### Examples

```
Copy a File:
./cp file1.txt file1_copy.txt

Recursive Directory Copy:
./cp -r dir1 dir2

Wildcard Copy:
./cp -v "*.txt" dest_dir

Verbose with Progress:
./cp -r -v large_dir dest_dir

Help:
./cp -h
```

```
Options:
  -r    Recursive copy (for directories)
  -i    Interactive mode (prompt before overwrite)
  -v    Verbose mode
  -p    Preserve file attributes (permissions and timestamps)
  -u    Update only (copy if source is newer than destination)
  -n    Dry-run mode (show actions without copying)
  -h    Display this help message
Notes: Supports wildcards (e.g., '*.txt') for source files.
```

## Testing
Run these tests in the project directory after setting up test files.


**Setup Test Environment**:  
   ```bash
   cd ~/cp-command-line-program
   mkdir test_cp
   cd test_cp
   echo "File 1 content" > file1.txt
   echo "File 2 content" > file2.txt
   echo "Another text" > file3.txt
   mkdir subdir
   echo "Subdirectory file" > subdir/subfile.txt
   mkdir large_dir
   echo "File A" > large_dir/fileA.txt
   echo "File B" > large_dir/fileB.txt
   echo "File C" > large_dir/fileC.txt
   touch -t 202301011200 oldfile.txt
   chmod 644 oldfile.txt
   echo "Old file content" > oldfile.txt
```
![Files setup](screenshots/Directory.png)

1. **Basic File Copy**:  

   ```bash
   ../cp file1.txt file1_copy.txt
   # Expected: No output; file1_copy.txt created.
   # Verify:
   cat file1_copy.txt  # Outputs: "File 1 content"

![Basic file copy](screenshots/Basic_file_copy.png)

2. **Verbose Mode (-v)**:  

   ```bash
   ../cp -v file1.txt file1_copy.txt
   # Expected:
   # Copied file "file1.txt" to "file1_copy.txt"
   # Verify:
   cat file1_copy.txt

   # Cleanup:
   rm file1_copy.txt

![verbose mode](screenshots/verbose_mode.png)

3. **Interactive Mode (-i)**:  

   ```bash
   cp file1.txt file1_copy.txt
   ../cp -i file1.txt file1_copy.txt
   # Expected:
   # File "file1_copy.txt" already exists. Overwrite? (y/n):
   # Verify:
   cat file1_copy.txt

   # Cleanup:
   rm file1_copy.txt

![Interactive mode](screenshots/Interactive_mode.png)

4. **Recursive Copy (-r)**:  

   ```bash
   ../cp -r -v subdir subdir_copy
   # Expected:
   # Created directory: "subdir_copy"
   # Copied file "subdir/subfile.txt" to "subdir_copy/subfile.txt"
   # Progress: Copied 1 files
   # Total files copied: 1
   # Verify:
   cat subdir_copy/subfile.txt

   # Cleanup:
   rm -r subdir_copy

![Recursive mode](screenshots/Recursive_copy.png)

5. **Preserve Attributes (-p)**:  

   ```bash
   ../cp -p -v oldfile.txt oldfile_copy.txt
   # Expected:
   # Copied file "oldfile.txt" to "oldfile_copy.txt"
   # Verify:
   ls -l oldfile.txt oldfile_copy.txt  # Timestamps and permissions match

   # Cleanup:
   rm oldfile_copy.txt

![preserve attribute](screenshots/preserve_attribute.png)

6. **Help Message (-h)**:  

   ```bash
   ./cp -h
   # Expected:
   # Usage: ./cp [options] source_file... destination
   # Options:
   #   -r    Recursive copy (for directories)
   #   -i    Interactive mode (prompt before overwrite)
   #   -v    Verbose mode
   #   -p    Preserve file attributes (permissions and timestamps)
   #   -u    Update only (copy if source is newer than destination)
   #   -n    Dry-run mode (show actions without copying)
   #   -h    Display this help message
   # Notes: Supports wildcards (e.g., '*.txt') for source files.

![help](screenshots/help.png)

7. **Multiple Files to Directory**:  

   ```bash
   mkdir dest_dir
   ../cp -v file1.txt file2.txt dest_dir
   # Expected:
   # Copied file "file1.txt" to "dest_dir/file1.txt"
   # Copied file "file2.txt" to "dest_dir/file2.txt"

   # Verify:
   ls dest_dir

   # Cleanup:
   rm -r dest_dir

![Multiple Files](screenshots/Multiple_files.png)

8. **Wildcard Support**:  

   ```bash
   mkdir dest_dir
    ../cp -v *.txt dest_dir
   # Expected:
   # Copied file "file1.txt" to "dest_dir/file1.txt"
   # Copied file "file2.txt" to "dest_dir/file2.txt"
   # Copied file "file3.txt" to "dest_dir/file3.txt"
   # Copied file "oldfile.txt" to "dest_dir/oldfile.txt"

   # Verify:
   ls dest_dir

   # Cleanup:
   rm -r dest_dir

![WildCard Support](screenshots/wildcard_support.png)

9. **Progress Indicator**:  

   ```bash
   ../cp -r -v large_dir large_copy

   # Expected:
   # Created directory: "large_copy"
   # Copied file "large_dir/fileA.txt" to "large_copy/fileA.txt"
   # Progress: Copied 1 files
   # Copied file "large_dir/fileB.txt" to "large_copy/fileB.txt"
   # Progress: Copied 2 files
   # Copied file "large_dir/fileC.txt" to "large_copy/fileC.txt"

   # Progress: Copied 3 files
   # Total files copied: 3

   # Verify:
   ls large_copy

   # Cleanup:
   rm -r large_copy

![progress indicator](screenshots/progress.png)

10. **Error Handling**

- **Non-existent Source**:  
   ```bash
   ../cp nonexistent.txt dest.txt

   # Expected:
   # Source file "nonexistent.txt" does not exist.


![Error Handeling](screenshots/error_handeling.png)

