#include "cp.hpp"
#include <iostream>
#include <experimental/filesystem>
#include <fstream>

// Alias for the experimental filesystem namespace
namespace fs = std::experimental::filesystem;

// Prompts the user to overwrite an existing file in interactive mode
// Returns true if the user agrees to overwrite, false otherwise
bool promptOverwrite(const std::string& filePath) {
    std::cout << "File \"" << filePath << "\" already exists. Overwrite? (y/n): ";
    char response;
    std::cin >> response;
    return (response == 'y' || response == 'Y');
}

// Copies a single file from source to destination with specified options
bool copyFile(const std::string& source, const std::string& destination,
              bool interactive, bool verbose, bool preserve, bool update, bool dryRun) {
    try {
        // Convert source and destination to filesystem paths
        fs::path srcPath(source);
        fs::path destPath(destination);

        // Check if the source file exists
        if (!fs::exists(srcPath)) {
            std::cerr << "Source file \"" << source << "\" does not exist.\n";
            return false;
        }

        // If destination is a directory, append the source filename
        if (fs::exists(destPath) && fs::is_directory(destPath)) {
            destPath /= srcPath.filename();
        }

        // In interactive mode, prompt before overwriting an existing file
        if (interactive && fs::exists(destPath)) {
            if (!promptOverwrite(destPath.string())) {
                if (verbose) std::cout << "Skipping file: " << destPath << "\n";
                return true;
            }
        }

        // In update mode, skip if destination is newer than source
        if (update && fs::exists(destPath)) {
            auto srcTime = fs::last_write_time(srcPath);
            auto destTime = fs::last_write_time(destPath);
            if (destTime >= srcTime) {
                if (verbose) std::cout << "Skipping file (destination is newer): " << destPath << "\n";
                return true;
            }
        }

        // In dry-run mode, simulate the copy without performing it
        if (dryRun) {
            if (verbose) {
                std::cout << "Dry run: would copy \"" << srcPath.string() << "\" to \""
                          << destPath.string() << "\"\n";
            }
            return true;
        }

        // Perform the actual file copy, overwriting if the destination exists
        fs::copy_file(srcPath, destPath, fs::copy_options::overwrite_existing);

        // Preserve file attributes if requested
        if (preserve) {
            fs::permissions(destPath, fs::status(srcPath).permissions());
            fs::last_write_time(destPath, fs::last_write_time(srcPath));
        }

        // Print verbose output if requested
        if (verbose) {
            std::cout << "Copied file \"" << srcPath.string() << "\" to \""
                      << destPath.string() << "\"\n";
        }
        return true;
    } catch (const fs::filesystem_error& e) {
        std::cerr << "Error copying file: " << e.what() << "\n";
        return false;
    }
}

// Recursively copies a directory and its contents from source to destination
// Parameters match those of copyFile; see above for descriptions
// Returns true on success, false on failure
bool copyDirectory(const std::string& sourceDir, const std::string& destinationDir,
                   bool interactive, bool verbose, bool preserve, bool update, bool dryRun) {
    try {
        // Convert source and destination to filesystem paths
        fs::path srcDir(sourceDir);
        fs::path destDir(destinationDir);

        // Validate that the source is an existing directory
        if (!fs::exists(srcDir) || !fs::is_directory(srcDir)) {
            std::cerr << "Source directory \"" << sourceDir << "\" does not exist or is not a directory.\n";
            return false;
        }

        // Create the destination directory if it doesn’t exist
        if (!fs::exists(destDir)) {
            if (dryRun) {
                if (verbose) std::cout << "Dry run: would create directory: " << destDir << "\n";
            } else {
                fs::create_directories(destDir);
                if (verbose) std::cout << "Created directory: " << destDir << "\n";
            }
        }

        int fileCount = 0; // Progress indicator
        for (const auto& entry : fs::recursive_directory_iterator(srcDir)) {
            const auto& path = entry.path();
            std::string relativePath = path.string().substr(srcDir.string().length());
            fs::path targetPath = destDir / relativePath;

            if (fs::is_directory(path)) {
                if (!fs::exists(targetPath)) {
                    if (dryRun) {
                        if (verbose) std::cout << "Dry run: would create directory: " << targetPath << "\n";
                    } else {
                        fs::create_directories(targetPath);
                        if (verbose) std::cout << "Created directory: " << targetPath << "\n";
                    }
                }
            } else if (fs::is_regular_file(path)) {
                fileCount++;
                copyFile(path.string(), targetPath.string(), interactive, verbose, preserve, update, dryRun);
                if (verbose && !dryRun) {
                    std::cout << "Progress: Copied " << fileCount << " files\n"; // New line instead of \r
                }
            }
        }
        // Print the total number of files copied
        if (verbose && !dryRun) {
            std::cout << "Total files copied: " << fileCount << "\n";
        }
        return true;
    } catch (const fs::filesystem_error& e) {
        std::cerr << "Error copying directory: " << e.what() << "\n";
        return false;
    }
}