#include <iostream>
#include <vector>
#include <string>
#include <experimental/filesystem>
#include <regex>
#include "cp.hpp"

// Alias for the experimental filesystem namespace
namespace fs = std::experimental::filesystem;


void printUsage(const std::string& programName) {
    std::cout << "Usage: " << programName << " [options] source_file... destination\n"
              << "Options:\n"
              << "  -r    Recursive copy (for directories)\n"
              << "  -i    Interactive mode (prompt before overwrite)\n"
              << "  -v    Verbose mode\n"
              << "  -p    Preserve file attributes (permissions and timestamps)\n"
              << "  -u    Update only (copy if source is newer than destination)\n"
              << "  -n    Dry-run mode (show actions without copying)\n"
              << "  -h    Display this help message\n"
              << "Notes: Supports wildcards (e.g., '*.txt') for source files.\n";
}

// Expands wildcard patterns into a list of matching files
// Parameters:
// - pattern: The wildcard pattern (e.g., "*.txt")
// Returns a vector of file paths that match the pattern

std::vector<std::string> expandWildcards(const std::string& pattern) {
    std::vector<std::string> matches;
    std::string dir = ".";
    std::string pat = pattern;
    if (pattern.find('/') != std::string::npos) {
        dir = pattern.substr(0, pattern.find_last_of('/'));
        pat = pattern.substr(pattern.find_last_of('/') + 1);
    }
    std::regex re(std::regex_replace(pat, std::regex("\\*"), ".*"));
    try {
        for (const auto& entry : fs::directory_iterator(dir)) {
            if (std::regex_match(entry.path().filename().string(), re)) {
                matches.push_back(entry.path().string());
            }
        }
    } catch (const fs::filesystem_error& e) {
        std::cerr << "Error expanding wildcard '" << pattern << "': " << e.what() << "\n";
    }
    return matches;
}

// Main function: Parses command-line arguments and initiates the copy process
// Parameters:
// - argc: Number of command-line arguments
// - argv: Array of command-line arguments
// Returns 0 on success, 1 on failure
int main(int argc, char* argv[]) {
    // Check for minimum number of arguments (at least source and destination)
    if (argc < 3) {
        printUsage(argv[0]);
        return 1;
    }

    // Initialize flags for copy options
    bool recursive = false, interactive = false, verbose = false;
    bool preserve = false, update = false, dryRun = false;
    std::vector<std::string> files;

    // Parse command-line arguments
    for (int i = 1; i < argc; ++i) {
        std::string arg = argv[i];
        // Handle option flags
        if (arg[0] == '-') {
            for (size_t j = 1; j < arg.size(); ++j) {
                switch (arg[j]) {
                    case 'r': case 'R': recursive = true; break;
                    case 'i': interactive = true; break;
                    case 'v': verbose = true; break;
                    case 'p': preserve = true; break;
                    case 'u': update = true; break;
                    case 'n': dryRun = true; break;
                    case 'h': printUsage(argv[0]); return 0;
                    default:
                        std::cerr << "Unknown flag: -" << arg[j] << "\n";
                        printUsage(argv[0]);
                        return 1;
                }
            }
        } else if (arg.find('*') != std::string::npos) {
            auto expanded = expandWildcards(arg);
            // Handle wildcard patterns
            if (expanded.empty()) {
                std::cerr << "No matches found for wildcard: " << arg << "\n";
            } else {
                files.insert(files.end(), expanded.begin(), expanded.end());
            }
        } else {
            files.push_back(arg);
        }
    }

    // Ensure at least one source and one destination are provided
    if (files.size() < 2) {
        std::cerr << "Error: Source and destination must be specified.\n";
        printUsage(argv[0]);
        return 1;
    }

    std::string destination = files.back();
    files.pop_back();

    // Check if the destination is a directory
    bool destIsDirectory = fs::exists(destination) && fs::is_directory(destination);
    if (files.size() > 1 && !destIsDirectory) {
        std::cerr << "Error: Destination must be a directory when copying multiple sources.\n";
        return 1;
    }

    // Process each source file or directory
    bool success = true;
    for (const auto& source : files) {
        fs::path srcPath(source);
        if (!fs::exists(srcPath)) {
            std::cerr << "Source \"" << source << "\" does not exist.\n";
            success = false;
            continue;
        }

        // Construct the destination path
        std::string destPath = destIsDirectory ? (destination + "/" + srcPath.filename().string()) : destination;
        if (fs::is_directory(srcPath)) {
            if (!recursive) {
                std::cerr << "Omitting directory \"" << source << "\" (use -r to copy directories).\n";
                success = false;
                continue;
            }
            if (!copyDirectory(source, destPath, interactive, verbose, preserve, update, dryRun)) {
                success = false;
            }
        } else if (fs::is_regular_file(srcPath)) {
            if (!copyFile(source, destPath, interactive, verbose, preserve, update, dryRun)) {
                success = false;
            }
        } else {
            // Handle unsupported file types
            std::cerr << "Unsupported file type: " << source << "\n";
            success = false;
        }
    }

    // Return 0 if all operations succeeded, 1 otherwise
    return success ? 0 : 1;
}